# streamlit_app.py

import streamlit as st
from langchain.vectorstores import FAISS
from langchain.embeddings import OpenAIEmbeddings
from langchain.chat_models import AzureChatOpenAI
from langchain.prompts import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
import pyodbc
import datetime
import os

# TODO: This is painfully modularizable.

# secrets
def manage_sensitive(name):
    v1 = os.getenv(name)
    secret_fpath = f'/run/secrets/{name}'
    existence = os.path.exists(secret_fpath)
    
    if v1 is not None:
        return v1
    
    if existence:
        v2 = open(secret_fpath).read().rstrip('\n')
        return v2
    
    if not existence:
        print(name)
        return KeyError(f'{name}')

DB_SERVER = manage_sensitive("db_server")
DB_NAME = manage_sensitive("db_name")
DB_USER = manage_sensitive("db_user")
DB_PASSWORD = manage_sensitive("db_password")
OPENAI_API_KEY = manage_sensitive("openai_api_key")

# openai
os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY

# db connection
def get_db_connection():
    conn_str = (
        "DRIVER={ODBC Driver 17 for SQL Server};SERVER="
        + DB_SERVER
        + ";DATABASE="
        + DB_NAME
        + ";UID="
        + DB_USER
        + ";PWD="
        + DB_PASSWORD
    )
    return pyodbc.connect(conn_str)

# page metadata
st.set_page_config(
    page_title="MPOG Helper ",
    page_icon="🤖",
)
# hide streamlit branding
hide_streamlit_style = """
            <style>
            #MainMenu {visibility: hidden;}
            footer {visibility: hidden;}
            </style>
            """
st.markdown(hide_streamlit_style, unsafe_allow_html=True) 

# page content
st.title("🗃️ MPOG Feasibility Checker 🤖")
st.markdown("""
**Determine if your research idea is appropriate for the data in MPOG**

Brought to you by the Anesthesiology MPOG Steering Committee, Informatics, and Data Science teams

_Not affiliated with MPOG. Not approved for use with sensitive data or PHI._

All submissions are recorded for potential review by the MPOG Steering Committee.

---
""")

st.sidebar.markdown("""
**Generative AIs are not reliable for factual information.**

They are helpful for synthesis and idea generation. They are predicting a response to what you say based on their training material (in this case pages from the internet).

You can trick generative AIs into saying whatever you want. If you succeed in tricking that AI, then what it says is something you created, not something it created.

As always, use your best judgment. An AI can do the wrong thing at the worst time.
""")

chat = AzureChatOpenAI(
    openai_api_base="https://nlp-openai-svc.openai.azure.com",
    openai_api_version="2023-03-15-preview",
    deployment_name="ChatGPT35Turbo",
    openai_api_key=OPENAI_API_KEY,
    openai_api_type = "azure",
    temperature=0
)
# Load vectorstore
embedding = OpenAIEmbeddings(
                deployment="AdaEmbedding2",
                model="text-embedding-ada-002",
                openai_api_base="https://nlp-openai-svc.openai.azure.com",
                openai_api_type="azure",
                openai_api_key=OPENAI_API_KEY,
                openai_api_version="2023-03-15-preview",
            )

vectordb = FAISS.load_local("faiss_index", embedding)
docsearch = vectordb.as_retriever()

with st.form(key="query_form"):
    user_question = st.text_input("Enter your research question:")
    submit_button = st.form_submit_button("Submit")

    if submit_button:
        docs = docsearch.get_relevant_documents(user_question)

        template="""You are a helpful assistant that tells a researcher whether their idea is appropriate given the data in the MPOG database.
        The MPOG database contains the following data elements that may be relevant to the researcher's ideas.
        {context}
        Additionally, patient demographics like age, race, gender, and BMI (body mass index) are available.
        Respond first by indicating whether the data in MPOG seems appropriate for the researcher’s idea. The first word in your response should be “Yes” or “No.” If their idea is not reasonable given the data in MPOG, explain what data elements that they might need are not in MPOG.

        If their idea is reasonable given the data in MPOG, briefly do the following 
        (1) Explain by mentioning (and suggesting they request) the specific data elements (but not table names) in MPOG.
        (2) Restate their research idea as a research hypothesis that uses the data in MPOG. 
        (3) Determine the novelty of their idea by comparing against the titles from the "MPOG publications" list below. If their idea is not closely matched by one of these titles, tell them their ideas seems novel. If their idea is closely matched by one of these titles, tell them their idea may not be novel, mention the matching article titles, and tell them they can find these articles on the "MPOG publications page."

        The following is a list of published articles using data in MPOG.
        *Prolonged Opioid Use and Pain Outcome And Associated Factors after Surgery Under General Anesthesia
        *Postoperative acute kidney injury by age and sex
        *Intraoperative Use of Albumin in Major non-cardiac surgery: Incidence, Variability, and Association with Outcomes
        *Oxygen administration during surgery and postoperative organ injury
        *Analysis of practice patterns regarding benzodiazepine use in cardiac surgery
        *Association of Anesthesiologist Staffing Ratio With Surgical Patient Morbidity and Mortality
        *Association between the choice of reversal agent for neuromuscular block and postoperative pulmonary complications in patients at increased risk undergoing non-emergency surgery
        *Assessment of Perioperative Outcomes Among Surgeons Who Operated the Night Before
        *Frequency and risk factors for difficult Intubation in women undergoing general anesthesia for cesarean delivery
        *Adherence to guidelines for the administration of intraoperative antibiotics in a Nationwide US sample
        *Hypoxemia in young children undergoing one-lung ventilation
        *Outcomes of surgical patients during the first wave of COVID-19 pandemic in US hospitals
        *Practice patterns and variability in intraoperative opioid utilization
        *Variation in propofol induction doses administered to surgical patients over age 65
        *A lower tidal volume regimen during one-lung ventilation for lung resection surgery is not associated with reduced postoperative pulmonary complications
        *Utilization patterns of perioperative neuromuscular blockade reversal in the United States
        *Sugammadex versus neostigmine for reversal of neruomuscular blockade and postoperative pulmonary complications
        *The incidence of intraoperative hypotension in moderate to high risk patients underoing non-cardiac surgery
        *Risk factors for intraoperative hypoglycemia in children
        *Postoperative pain profies, analgesic use, and transition to chronic pain and excessive prologed opioid use patterns methodology
        *Considerations for Integration of Perioperative Electronic Health Records Across Institutions for Research and Quality Improvement
        *Classification of current procedural terminology codes from electronic health record data using machine learning
        *Periopertive risk and the association between hypotension and postoperative acute kidney injury
        *Making sense of big data to improve perioperative care: Learning health systems
        *Improved Compliance With Anesthesia Quality Measures After Implementation of Automated Monthly Feedback
        *Association of overlapping surgery with perioperative outcomes
        *An observational study of end-tidal carbon dioxide trends in general anesthesia
        *Succinylcholine use and dantrolene availability for malignant hyperthermia treatment: Database analyses and systematic review
        *Management of 1-lung ventilator – variation and trends in clinical practic
        *Alarm limits for intraoperative drug infusions
        *Risk of epidural hematoma after neuraxial techniques in thrombocytopenic parturients
        *Reference values for noninvasive blood pressure in children during anesthesia
        *Success of intubation rescue techniques after failed direct laryngoscopy in adults
        *Intraoperative lung-protection ventilation treds and practice patterns
        *Incidence, predictors, and outcomes of difficult mask ventilation combined with difficult laryngoscopy
        *The risk and outcomes of epidural hematomas after perioperative and obstetric epidural catheterization
        *Perioperative effectiveness of research using large databases
        *Clinical research using an information system

        Do not consider any information outside of this list of article titles when determining novelty!

        Phrase your response to the researcher directly.
        """

        system_message_prompt = SystemMessagePromptTemplate.from_template(template)
        human_template = """My research idea is: {question}"""
        human_message_prompt = HumanMessagePromptTemplate.from_template(human_template)
        chat_prompt = ChatPromptTemplate.from_messages(
            [system_message_prompt, human_message_prompt])
        with st.spinner("Thinking..."):
            submit_time = datetime.datetime.now()
            result = chat(chat_prompt.format_prompt(context=docs, question=user_question).to_messages())
            response_time = datetime.datetime.now()
        st.markdown(result.content)
        
        try:
            with get_db_connection() as conn:
                cursor = conn.cursor()
                query = """
                INSERT INTO [dbo].[mpog_helper] (user_input, llm_response, request_sent, response_received)
                VALUES (?, ?, ?, ?)
                """

                cursor.execute(query, (user_question, result.content, submit_time, response_time))

                st.success("Your idea has also been recorded and may be reviewed by the MPOG Steering Committee.")
        except Exception as e:
            st.error("Something went wrong, and your idea was not recorded for review by the MPOG Steering Committee. Give the following message when asking for help.")
            st.error(e)
